# Copyright (c) 2023 Zeeland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/cushy-storage
# Contact Email: zeeland@foxmail.com

import os
import tempfile
import platform


def get_project_root_path() -> str:
    """get project root path"""
    project_path = os.getcwd()
    max_depth = 10
    count = 0
    while not os.path.exists(os.path.join(project_path, "README.md")):
        project_path = os.path.split(project_path)[0]
        count += 1
        if count > max_depth:
            return os.getcwd()
    return project_path


def get_default_storage_path(file_name) -> str:
    if platform.system() == "Windows":
        return f"./{file_name}"
    elif platform.system() == "Linux" or "Darwin":
        dir_path = os.environ.get("TMPDIR")
        if not dir_path:
            dir_path = tempfile.gettempdir()
        dir_path = os.path.join(dir_path, "cushy_storage")
        return f"{dir_path}/{file_name}"
    else:
        return f"./{file_name}"


def get_default_cache_path() -> str:
    return get_default_storage_path("cache")


def get_default_log_path() -> str:
    return get_default_storage_path("log")



# Copyright (c) 2023 Zeeland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/cushy-storage
# Contact Email: zeeland@foxmail.com

import os
import zlib
import lzma
import json
import pickle
import hashlib
import threading
from pathlib import Path
from typing import MutableMapping, Callable, Tuple, Union, Any, List

from cushy_storage.base import EnhancedList, BASE_TYPE
from cushy_storage.utils import get_default_cache_path

__all__ = ["BaseDict", "CushyDict", "disk_cache"]

# Compression algorithms and their corresponding functions
_COMPRESS = {
    "zlib": (
        zlib.compress,
        zlib.decompress,
    ),
    "lzma": (
        lzma.compress,
        lzma.decompress,
    ),
}

# Serialization algorithms and their corresponding functions
_SERIALIZATION = {
    "pickle": (
        pickle.dumps,
        pickle.loads,
    ),
    "json": (
        lambda x: json.dumps(
            x, sort_keys=True, ensure_ascii=False, separators=(",", ":")
        ).encode("utf8"),
        json.loads,
    ),
}

# Locks for each hash value (hexadecimal representation of 0-255)
_LOCKS = {hex(i)[2:].zfill(2): threading.Lock() for i in range(256)}


def _method_convert_helper(
        s: Union[str, Tuple[Callable, Callable], None], d: dict
) -> Tuple[Callable, Callable]:
    """
    Helper function to get the compression or serialization functions based on input parameter
    """
    if s is None:
        return lambda x: x, lambda x: x
    elif isinstance(s, str):
        return d[s]
    else:
        return s


class BaseDict(MutableMapping[str, bytes]):
    def __init__(
            self, path: str, compress: Union[str, Tuple[Callable, Callable], None] = None
    ):
        self.path = Path(path)
        if self.path.is_file():
            raise Exception(
                "path has exist"
            )  # Raise an exception if the path already exists as a file
        self.path.mkdir(parents=True, exist_ok=True)
        self.dirs = set()
        self.compress, self.decompress = _method_convert_helper(compress, _COMPRESS)

    def __contains__(self, k: str):
        """
        Check if the file exists in the cache

        Args:
            k: key

        Examples:
            from cushy_storage import CushyDict

            cache = CushyDict()
            if 'my_key' in cache:
                print("[my_key] in my cache")
            else:
                print("[my_key] not in my cache")
        """
        return (self.path / k[:2] / (k[2:] + "_")).is_file()

    def __getitem__(self, k: str):
        """
        Retrieve the cached item using its key and decompress it
        """
        if k not in self:
            raise KeyError(k)
        rk = hashlib.md5(k.encode("utf8")).hexdigest()[:2]
        with _LOCKS[rk]:
            with open(self.path / k[:2] / (k[2:] + "_"), "rb") as f:
                t = f.read()
        return self.decompress(t)

    def __setitem__(self, k: str, v: bytes):
        """
        Compress the value and store it in the cache using its key
        """
        if k[:2] not in self.dirs:
            (self.path / k[:2]).mkdir(exist_ok=True)
            self.dirs.add(k[:2])
        t = self.compress(v)
        rk = hashlib.md5(k.encode("utf8")).hexdigest()[:2]
        with _LOCKS[rk]:
            with open(self.path / k[:2] / (k[2:] + "_"), "wb") as f:
                f.write(t)

    def __delitem__(self, k: str):
        """
        Remove the cached item using its key
        """
        os.remove(self.path / k[:2] / (k[2:] + "_"))

    def __len__(self):
        """
        Get the total number of items in the cache
        """
        return sum([len(os.listdir(self.path / a)) for a in os.listdir(self.path)])

    def __iter__(self):
        """
        Iterate over all keys in the cache
        """
        for a in os.listdir(self.path):
            for b in os.listdir(self.path / a):
                yield a + b[:-1]


class CushyDict(BaseDict):
    """
    A subclass of BaseDict that can serialize and deserialize values using different algorithms
    """

    def __init__(
            self,
            path: str = get_default_cache_path(),
            compress: Union[str, Tuple[Callable, Callable], None] = None,
            serialize: Union[str, Tuple[Callable, Callable], None] = "json",
    ):
        """
        Args:
            path: the path to save
            compress: zlib or lzma
            serialize: json or pickle
        """
        super().__init__(path, compress)
        self.serialize, self.deserialize = _method_convert_helper(
            serialize, _SERIALIZATION
        )

    def __getitem__(self, k: str) -> Any:
        ret = self.deserialize(super().__getitem__(k))

        if isinstance(ret, list):
            ret: List = EnhancedList(ret)
        return ret

    def __setitem__(self, k: str, v: Any):
        if (
                isinstance(v, list)
                and self.deserialize is json.loads
                and len(v) > 0
                and type(v[0]) not in BASE_TYPE
        ):
            raise ValueError(
                (
                    f"Can not use 'json' to serialize your '{type(v[0])}' data in '{k}'. If you want to store "
                    "complex data or custom data, please use 'pickle' to serialize."
                )
            )
        return super().__setitem__(k, self.serialize(v))


def disk_cache(path: str = None, compress: str = None, serialize: str = "json"):
    """
    Decorator that caches the output of a function to disk
    """
    if serialize not in ["pickle", "json"]:
        ValueError("Your serializer must be 'pickle' or 'json'")
    dump = _method_convert_helper(serialize, _SERIALIZATION)[0]

    def decorator(func):
        nonlocal path
        name = func.__name__
        if path is None:
            # If no cache path is specified, create a default one based on the function name and serialization algorithm
            path = f"./_cushycache_{name}_{serialize}"
        map = CushyDict(path, serialize=serialize, compress=compress)

        def cached_func(*args, **kwargs):
            # Serialize the function arguments and use their MD5 hash as the cache key
            input_data = [name, args, kwargs]
            md5 = hashlib.md5(dump(input_data)).hexdigest()
            ext = "pkl" if serialize == "pickle" else "json"
            filename = f"{md5}.{ext}"

            if filename in map:
                # If the cached output exists, return it
                input_data, output_data = map[filename]
                return output_data
            else:
                # Otherwise, call the original function and cache its output
                output_data = func(*args, **kwargs)
                cache_data = [input_data, output_data]
                map[filename] = cache_data
                return output_data

        return cached_func

    return decorator



# Copyright (c) 2023 Zeeland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/cushy-storage
# Contact Email: zeeland@foxmail.com

import os
import logging
import datetime
import platform
from cushy_storage import utils

logger = logging.getLogger("cushy-storage")


def get_logger():
    return logger


def _check_log_path():
    """check whether log file exist"""
    if not os.path.exists(utils.get_default_log_path()):
        os.makedirs(utils.get_default_log_path())


def get_log_name() -> str:
    _check_log_path()
    cur_time = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    return f"{utils.get_default_log_path()}/log_{cur_time}.log"


def enable_log():
    """enable logging to terminal and file"""
    logging.basicConfig(
        level=logging.DEBUG,
        format="[%(levelname)s] %(asctime)s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        handlers=[
            logging.FileHandler(f"{get_log_name()}", mode="w", encoding="utf-8"),
            logging.StreamHandler(),
        ],
    )


def enable_log_no_file():
    logging.basicConfig(
        level=logging.DEBUG,
        format="[%(levelname)s] %(asctime)s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )



# Copyright (c) 2023 Zeeland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/cushy-storage
# Contact Email: zeeland@foxmail.com

import json
import uuid
import hashlib
from abc import ABC
from typing import Union, Tuple, Callable, List, Optional

from cushy_storage import CushyDict
from cushy_storage.utils import get_default_cache_path
from cushy_storage.logger import get_logger

logger = get_logger()


class BaseORMModel(ABC):
    def __init__(self):
        self.__name__ = type(self).__name__
        self.__unique_id__: str = str(uuid.uuid4())

    def __get_element_hash__(self) -> str:
        dic = self.__dict__.copy()
        del dic["__unique_id__"]

        json_data = json.dumps(dic, sort_keys=True).encode("utf-8")
        hash256 = hashlib.sha256()
        hash256.update(json_data)
        return hash256.hexdigest()


class QuerySet:
    def __init__(
        self, obj: Union[List[BaseORMModel], BaseORMModel], name: Optional[str] = None
    ):
        self._data: List[BaseORMModel] = obj
        if isinstance(obj, BaseORMModel):
            self._data = [obj]
        if len(self._data) == 0 and name:
            self.__name__ = name
        else:
            self.__name__ = name if name else self._data[0].__name__

    @classmethod
    def _from_filter(
        cls, obj: Union[List[BaseORMModel], BaseORMModel], name: Optional[str] = None
    ) -> "QuerySet":
        """generate a new queryset from filter"""
        return cls(obj, name)

    def filter(self, **kwargs) -> "QuerySet":
        """
        filter by specified parameter
        Args:
            **kwargs: The property of the object you want to query

        Returns: return a new QuerySet object

        Examples:
            class User(BaseORMModel):
                def __init__(self, name, age):
                    super().__init__()
                    self.name = name
                    self.age = age
            # get all user, you will get a List[User] type data.
            # Actually, it will get two users named "jack" and "jasmine".
            orm_cache.query("User").filter(age=18).all()
            # get first in queryset, you will get a User type data
            orm_cache.query("User").filter(name="jack").first()
            # filter by multiple parameters
            orm_cache.query("User").filter(name="jack", age=18).first()
        """
        result: List[BaseORMModel] = []
        for item in self._data:
            is_target = True
            for query_key in kwargs.keys():
                if item.__dict__[query_key] != kwargs[query_key]:
                    is_target = False
                    continue
            if is_target:
                result.append(item)

        return self._from_filter(result, self.__name__)

    def all(self) -> Optional[List[BaseORMModel]]:
        return self._data

    def first(self) -> Optional[BaseORMModel]:
        if len(self._data) == 0:
            return None
        return self._data[0]

    def print_all(self):
        for item in self._data:
            print(f"[cushy-storage orm] {item.__dict__}")

    @classmethod
    def _from_remove_duplicates(
        cls, obj: List[BaseORMModel], name: Optional[str] = None
    ) -> "QuerySet":
        return cls(obj, name)

    def remove_duplicates(self) -> Optional["QuerySet"]:
        if len(self._data) == 0:
            return None

        result_element_hash = []
        result: List[BaseORMModel] = []
        for obj in self._data:
            if obj.__get_element_hash__() not in result_element_hash:
                result_element_hash.append(obj.__get_element_hash__())
                result.append(obj)
        return self._from_remove_duplicates(result, name=self.__name__)


def _get_class_name(class_name_or_obj: Union[str, type(BaseORMModel)]) -> str:
    class_name: str = class_name_or_obj
    if isinstance(class_name_or_obj, type(BaseORMModel)):
        class_name = class_name_or_obj.__name__
    return class_name


def _get_obj_name(obj: Union[BaseORMModel, QuerySet, List[BaseORMModel]]) -> str:
    if isinstance(obj, BaseORMModel) or isinstance(obj, QuerySet):
        return obj.__name__
    elif isinstance(obj, List):
        return obj[0].__name__


class ORMMixin(ABC):
    def _get_original_data_from_cache(
        self, class_name_or_obj: Union[str, type(BaseORMModel)]
    ) -> List[BaseORMModel]:
        class_name = _get_class_name(class_name_or_obj)
        if class_name not in self:
            self.__setitem__(class_name, [])
        return self.__getitem__(class_name)

    def query(self, class_name_or_obj: Union[str, type(BaseORMModel)]) -> QuerySet:
        """query all objects by class name"""
        original_result = self._get_original_data_from_cache(class_name_or_obj)
        if len(original_result) == 0:
            return QuerySet(original_result, name=_get_class_name(class_name_or_obj))
        return QuerySet(original_result)

    def remove_duplicates(self, class_name_or_obj: Union[type(BaseORMModel), str]):
        original_result = self._get_original_data_from_cache(class_name_or_obj)
        if len(original_result) != 0:
            queryset = QuerySet(original_result)
            queryset = queryset.remove_duplicates()
            self.set(queryset)

    def add(self, obj: Union[BaseORMModel, QuerySet, List[BaseORMModel]]) -> QuerySet:
        obj_name = _get_obj_name(obj)
        original_result: List[BaseORMModel] = self._get_original_data_from_cache(
            obj_name
        )

        if isinstance(obj, BaseORMModel):
            original_result.append(obj)
        elif isinstance(obj, QuerySet):
            original_result += obj.all()
        else:
            original_result += obj

        self[obj_name] = original_result
        return QuerySet(self[obj_name])

    def delete(self, obj: Union[List[BaseORMModel], QuerySet, BaseORMModel]):
        """delete obj by obj.__unique_id__"""
        obj_name = _get_obj_name(obj)
        original_result: List[BaseORMModel] = self._get_original_data_from_cache(
            obj_name
        )
        copy_result: List[BaseORMModel] = original_result.copy()

        if isinstance(obj, BaseORMModel):
            obj = [obj]
        elif isinstance(obj, QuerySet):
            obj = obj.all()
        if len(obj) == 0:
            return

        for cache_obj_item in original_result:
            for input_obj_item in obj:
                if cache_obj_item.__unique_id__ == input_obj_item.__unique_id__:
                    copy_result.remove(cache_obj_item)
        return self.__setitem__(obj_name, copy_result)

    def set(self, obj: Union[BaseORMModel, QuerySet, List[BaseORMModel]]):
        obj_name = _get_obj_name(obj)
        if isinstance(obj, BaseORMModel):
            obj = [obj]
        elif isinstance(obj, QuerySet):
            obj = obj.all()
        if len(obj) == 0:
            return
        return self.__setitem__(obj_name, obj)

    def update_obj(self, obj: BaseORMModel):
        original_result: List[BaseORMModel] = self._get_original_data_from_cache(
            obj.__name__
        )
        copy_result: List[BaseORMModel] = original_result.copy()

        for i in range(len(copy_result)):
            if copy_result[i].__unique_id__ == obj.__unique_id__:
                copy_result[i] = obj
                return self.__setitem__(obj.__name__, copy_result)

        raise ValueError(f"can not found object: {obj}")

    def __getitem__(self, item) -> List[BaseORMModel]:
        """implemented by CushyDict"""

    def __setitem__(self, key, value) -> List[BaseORMModel]:
        """implemented by CushyDict"""


class CushyOrmCache(CushyDict, ORMMixin):
    def __init__(
        self,
        path: str = get_default_cache_path(),
        compress: Union[str, Tuple[Callable, Callable], None] = None,
    ):
        super().__init__(path, compress, "pickle")



# Copyright (c) 2023 Zeeland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/cushy-storage
# Contact Email: zeeland@foxmail.com

import unittest
from cushy_storage import BaseDict


class TestBaseDict(unittest.TestCase):
    def test_basic_operations(self):
        cache = BaseDict("./cache/test-base-dict")

        # Test adding items to the cache
        cache["foo"] = b"bar"
        self.assertEqual(cache["foo"], b"bar")

        # Test deleting items from the cache
        del cache["foo"]
        with self.assertRaises(KeyError):
            cache["foo"]

        # Test checking if an item is in the cache
        self.assertFalse("foo" in cache)

    def test_compression(self):
        cache = BaseDict("./cache/test-base-dict", compress="lzma")

        # Test storing and retrieving a large string in the cache using LZMA compression
        data = "a" * (1024 * 1024)
        cache["big_data"] = data.encode()
        self.assertEqual(cache["big_data"].decode(), data)



# Copyright 2022 Zeeland(https://github.com/Undertone0809/). All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import setuptools
import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")

setuptools.setup(
    name="cushy-storage",
    version="1.2.2",
    author="Zeeland",
    author_email="zeeland@foxmail.com",
    description="A data local persistence ORM framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Undertone0809/cushy-storage",
    packages=setuptools.find_packages(),
    license="Apache 2.0",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords="storage, ORM, serialization, json, cushy-storage, cushy_storage",
)
